#!/usr/bin/perl
#-----------------------------------------------------------------------
# Licensed Materials - Property of IBM
#
# IBM Tape System Library Manager
#
# (C) Copyright IBM Corp. 2004, 2017 All Rights Reserved
#-----------------------------------------------------------------------
#
# @version 1.1 - 17/10/17 00:53:20
#
# @brief
#
# Change activities:
#     - 2017-10-30: Initial version
#
#-----------------------------------------------------------------------
package Ermm::CLI::LsClients;
use Getopt::Long;
use Term::ReadLine;
use Ermm::CLI::Conf;
use Ermm::CLI::Main;

$synopsis = "lsclients -host <hostip> [ -help, -? | -v | -x | -fmt [xml|stanza|delim=c]]";

#
# Change drive manager
#
sub lsclients {
    my ($i, $result, $host, $number, $response, $ermmOutput );
    my $help        = 0;
    my $sendString  = "";
    my $matchString = "";

    #
    # get defaults from configuration
    #
    my $format = $Ermm::CLI::Conf::format;
    my $delimiter = $Ermm::CLI::Conf::delimiter;
    my $header = $Ermm::CLI::Conf::header;

    #
    # the classname for xml output
    #
    my $className = "DM";

    #
    # attribute names on headline
    #
    my @propNames = ("Component", "Version");

    #
    # attribute types
    #
    my @typeNames = ("string", "string");

    #
    # process command line options
    #
    $result = GetOptions ("v"        => \$Ermm::CLI::Conf::verbose, # flag
                          "help|?"   => \$help,                     # flag
			              "x"        => \$ermmOutput,               # flag
                          "v"        => \$Ermm::CLI::Conf::verbose, # flag
                          "fmt=s"    => \$format,                   # string
                          "hdr=s"    => \$header,                   # string
                          "host=s"   => \$host);                    # string

    #
    # check if we should print usage infos
    #
    if ($result != 1 || $help == 1) {
        Ermm::CLI::Main::printtext('_USAGE_LSCLIENTS', $synopsis);
        exit 1;
    }

    #
    # check header option
    #
    if ($header eq "on" || $header eq "yes") {
        $header = 1;
    }

    #
    # get delimiter from -fmt option
    #
    if ($format =~ /^delim=(.*)/) {
        $format = "delim";
        $delimiter = $1;
    }

    #
    # ckeck if there are any arguments
    #
    unless ($host){
        Ermm::CLI::Main::printtext('EMM7010E');
        Ermm::CLI::Main::printtext('_USAGE_LSCLIENTS', $synopsis);
        exit 1;
    }
    my $chk=checkDMVersion($host);
    if ($chk==1){
	print "HDM component at host[$host] doesn't support this command. Please update the HDM to version 1.2.0.2 or later.\n";
	exit 1;
    }

    if ($host) {
        $matchString = "and($matchString streq(DM.\"DMHost\" ".
            "\"$host\"))";
    }

    if (length($matchString) > 0) {
        $matchString = "match[$matchString]";
    } else {
        Ermm::CLI::Main::printtext('EMM7010E');
        Ermm::CLI::Main::printtext('_USAGE_LSCLIENTS', $synopsis);
        exit 1;
    }

    if ($host){
	$sendString = "cpattribute task[\"$Ermm::CLI::Conf::taskId\"] cptype[dm] cpreport[\"clientsver\"]";
    }

    $sendString .= " $matchString;\n";

    #
    # connect to TSLM server
    #
    Ermm::CLI::Main::connect();

    #
    # send command to server
    #
    Ermm::CLI::Main::send($sendString);

    #
    # receive response from server
    #
    $response = Ermm::CLI::Main::getResponse();

    if ($ermmOutput) {
	#
	# -x option specified; print raw TSLM output
	#
	print $response,"\n";
    }
    else {
	#
	# start normal output processing
	#
	@values = Ermm::CLI::Main::getKeyValueResultsFromSingleText($response);
	
	if ($#values < 0) {
	    exit;
	}
	Ermm::CLI::Main::printResults($format, $delimiter, $header, $className,
                                      \@propNames, \@typeNames, \@values);
    }
}

sub checkDMVersion {
    # check
    # 1.2.0.2 or later (1.2.X.X)
    # is not 1.3.2.X
    # 1 reject
    # 0 accept

    my $host = shift;
    my $matchString = "and(streq(DM.\"DMHost\" \"$host\"))";
    my $orderString = "strlohi(DM.\"DMName\")";
    my $numberString = "1";
    my $reportString = "DM.\"ClientVersion\"";

    #
    # connect to TSLM server
    #
    Ermm::CLI::Main::connect();

    #
    # send command to server
    #
    Ermm::CLI::Main::sendMMPShow($Ermm::CLI::Conf::taskId,
                                 $matchString,
                                 $orderString,
                                 $numberString,
                                 $reportString,
                                 "");

    #
    # receive response from server
    #
    $response = Ermm::CLI::Main::getResponse();
    my $res = $response;
    $res =~ s/.*text\s*\[\"//g;
    $res =~ s/\"\]\s*$//g;
    my @ver= (split /\./, $res);

    #print "@ver[0],@ver[1],@ver[2],@ver[3]\n";

    if ( @ver[0]==1 and @ver[1]==3 and @ver[2]==2 ){
	return 1;
    }

    if ( @ver[0] > 1 ){
	return 0;
    }

    if ( @ver[1] > 2 ){
	return 0;
    }

    if ( @ver[2] > 0 ){
	return 0;
    }

    if ( @ver[3] >= 2 ){
	return 0;
    }
    return 1;
}


1;
