#!/usr/bin/perl
#-----------------------------------------------------------------------
# Licensed Materials - Property of IBM
#
# IBM Tape System Library Manager
#
# (C) Copyright IBM Corp. 2004, 2014 All Rights Reserved
#-----------------------------------------------------------------------
#
# @version 1.6 - 14/03/24 16:23:38
#
# @brief
#
#-----------------------------------------------------------------------

package ErmmHdm;
use 5.005;
use IO::Socket;
use IO::Select;
use constant PING => do{ eval { require Net::Ping }};
use constant IPV6 => do{ eval { require Socket::Class }};
use HDMVersion;
use genLogFiles;
use nls::hdm_L10N;

my $secondsBeforeExit = 60;
my $saveCommand       = "";
my $socket;
my $socket_IPV6;
my $select;

#
# Connect to TSLM server using the credentials.
#
# In case there are multiple IP addresses iterate through the list
# in order to find a running Media Manager.
#
# Returns: Local IP address used for the TCP connection.
#
sub connect {
    my ($host, $port, $client, $instance, $password, $proto,
        $sec, $locaddr, $timeout, $configFile, $ipv6) = @_;

    my ($line, $IPAddress);
    my @ipAddresses = ();
    my $numberOfIps = 0;
    my $index       = 0;
    my $reorder     = 0;
    my $foundIP     = 0;
    my $IPPort      = "";
    my $ipaddress   = "";

    if ($sec) {
        $secondsBeforeExit = $sec;
    }
    if ($ipv6) {
        $ipv6support = $ipv6;
    }
    else {
        $ipv6support = 0;
    }

    # Check to see if there is a comma separated list of IP addresses
    @ipAddresses = split(/,/,$host);
    $numberOfIps = $#ipAddresses+1;

    foreach my $currentIp (@ipAddresses) {

        ($IPAddress, $IPPort) = parsingIpAddress($currentIp, $port);

        if ($ipv6support && !IPV6) {
            genLogFiles::logprint("Socket::Class missing for IPv6.");
        }

        #
        # IPv6 connection
        #
        if ($ipv6support && IPV6) {

           # create a new socket and connect to the IPv6 server
           if (defined $locaddr && $locaddr ne "") {
              $socket_IPV6 = Socket::Class->new(
                                     'remote_addr' => $IPAddress,
                                     'remote_port' => $IPPort,
                                     'local_addr'  => $locaddr,
                                     'domain'      => "inet6",
              );
           }
           else {
              $socket_IPV6 = Socket::Class->new(
                                     'remote_addr' => $IPAddress,
                                     'remote_port' => $IPPort,
                                     'domain'      => "inet6",
              );
           }

           if (Socket::Class->is_error ||
               Socket::Class->error)   {
              if ($numberOfIps == 1) {
                 genLogFiles::logprint("cannot connect to IPv6 host $IPAddress on port $IPPort: $!");
                 doExit();
              }
              elsif ($numberOfIps > 1) {
                 genLogFiles::logprint("cannot connect to IPv6 host $IPAddress on port $IPPort: $!");
                 genLogFiles::logprint("Trying to connect using alternate IP address.");
                 $reorder=1;
                 next;
              }
           }

           my $tcpnodelay = $socket_IPV6->set_tcp_nodelay(1);
           if (!defined $tcpnodelay) {
              genLogFiles::logprint("can't set TCP_NODELAY for \'$IPAddress:$IPPort\': $!");
           }

           if (!defined $socket_IPV6->remote_addr) {
              $ipaddress = $IPAddress;
           }
           else {
              $ipaddress = $socket_IPV6->remote_addr;
           }
        }

        #
        # IPv4 connection
        #
        else {

           # Check if a given address is reachable on the network.
           # The check is only done if there is a timeout value greater than 0.
           # This mechanism is faster than waiting for the socket operation timeout
           # (or modifying any socket options).
           if (PING && ($timeout != 0)) {
              my $pingHost = Net::Ping->new();

              # Remark: $pingHost->port_number($IPPort) is only availbale with version 2.33 (July 2007)
              # Since this is not the default version we don't use this option (yet)
              if (!($pingHost->ping($IPAddress, $timeout))) {
                  genLogFiles::logprint("IP address $IPAddress was not reachable on the network ".
                                        "within $timeout seconds.");
                  $pingHost->close();

                  if ($numberOfIps == 1) {
                      doExit();
                  }
                  elsif ($index < $numberOfIps) {
                      genLogFiles::logprint("Trying to reach media manager on alternate IP address.");
                      $reorder=1;
                      next;
                  }
              }
              $pingHost->close();
           }

           if (defined $locaddr && $locaddr ne "") {
              #
              # create a socket
              #
              $socket = IO::Socket::INET->new(Proto     => "tcp",
                                              LocalAddr => $locaddr,
                                              PeerAddr  => $IPAddress,
                                              PeerPort  => $IPPort);
           }
           else {
              #
              # create a socket
              #
              $socket = IO::Socket::INET->new(Proto    => "tcp",
                                              PeerAddr => $IPAddress,
                                              PeerPort => $IPPort);
           }

           if (!$socket && ($numberOfIps == 1)) {
              genLogFiles::logprint("cannot connect to host \'$IPAddress:$IPPort\': $!");
              doExit();
           }
           elsif (!$socket && ($numberOfIps > 1)) {
              genLogFiles::logprint("cannot connect to host \'$IPAddress:$IPPort\': $!");
              genLogFiles::logprint("Trying to connect using alternate IP address.");
              $reorder=1;
              next;
           }

	   $socket->autoflush();  #74582

           my $tcpnodelay = setsockopt($socket,
                                       Socket::IPPROTO_TCP,
                                       Socket::TCP_NODELAY,
                                       pack("I",1));
           if (!$tcpnodelay) {
              genLogFiles::logprint("can't set TCP_NODELAY for \'$IPAddress:$IPPort\': $!");
           }

           $select = new IO::Select($socket);

           $ipaddress = inet_ntoa($socket->sockaddr());
        }

        genLogFiles::logprint("starting session with client=$client, instance=$instance, ".
                              "password=******");

        #
        # start session and authenticate ourself
        #
        my $hello_string = "";
        if ($HDMVersion::HDMVERSION eq "HDMVERSIONSTRING") {
            $hello_string = "hello client[\"$client\"] ".
                " instance[\"$instance\"] ".
                " password[\"$password\"] ".
                " language[\"$proto\"];\n";
        }
        else {
            $hello_string = "hello client[\"$client\"] ".
                " instance[\"$instance\"] ".
                " password[\"$password\"] ".
                " language[\"$proto\"] clientversion [\"$HDMVersion::HDMVERSION\"];\n";
        }

        ErmmHdm::send($hello_string);
        $line = getWellcome();

        if (!defined $line && ($numberOfIps == 1)) {
            genLogFiles::logprint("unable start session with client=\"$client\", ".
                                  "instance=\"$instance\", password=\"******\"");
            closeSocket();
            doExit();
        }
        elsif (!defined $line && ($numberOfIps > 1)) {
            genLogFiles::logprint("unable start session with client=\"$client\", ".
                                  "instance=\"$instance\", password=\"******\"");
            genLogFiles::logprint("Trying to connect using alternate IP address.");
            closeSocket();
            $reorder=1;
            next;
        }

        if ((!defined $line || $line =~ /unwelcome/) && ($numberOfIps == 1)) {
            genLogFiles::logprint("got unwelcome response: ", $line);
            closeSocket();
            doExit();
        }
        elsif ((!defined $line || $line =~ /unwelcome/) && ($numberOfIps > 1)) {
            genLogFiles::logprint("got unwelcome response: ", $line);
            genLogFiles::logprint("Trying to connect using alternate IP address.");
            closeSocket();
            $reorder=1;
            next;
        }
        else {
            # ensure we stop foreach loop iterarion if we have a valid connection
            $foundIP = 1;
        }

        if ($foundIP) { last };

    } # end of foreach

    # Only reorder/rewrite the IP list in the configuration file if there is a list of
    # IP address and an address other than the first was reachable on the network.
    if (($numberOfIps>1) && $reorder && $foundIP) {
        reorderIPs($IPAddress, $IPPort, $host, $port, $configFile);
    }

    # In case no IP address was reachable
    if (!$foundIP) {
        doExit();
    }

    # Everything went well
    return $ipaddress;
}

#
# send a command string to the TSLM server
#
sub send {
    my $line = shift;

    genLogFiles::logprint("sending =>$line<=");

    if ($ipv6support && IPV6) {
        $bytes = $socket_IPV6->send($line);
        if (!$bytes) {
            genLogFiles::logprint("Error during send : ". $socket_IPV6->error);
            doExit();
        }
    }
    else {
	 print $socket $line;
	 $socket->flush(); #74582
#        syswrite($socket,$line);
    }
}

#
# send message to Media Manager
#
# Arguments: Message id followed by severity, then key, value pairs for the arguments,
#            e.g. sendMessage(EMM1234, "DMHost", "127.0.0.1", "SerialNumber", "0815")
#
sub sendMessage {
  my ($id, $severity, @args) = @_;

  my $cmd = qq/message task ["hdmmsg"] who ["DM"] severity ["$severity"] message [ id [ "IBM" "TSLM" "$id" ] arguments [ /;
  foreach (@args) {
    $cmd .= qq/"$_" /;
  }
  $cmd .= qq/ ] ] ;\n/;

  ErmmHdm::send($cmd);
}

#
# send sense data via config cmd to Media Manager
#
sub sendDriveSenseData {
    my ($drvsnr, $sd) = @_;
    my $cmd = qq/config task ["hdmsense"] scope["full"] dm["key" "SerialNumber" "SerialNumber" "$drvsnr"] drivesense["$sd"];\n/;

    ErmmHdm::send($cmd);
}


#
# disconnect from the TSLM server
#
sub disconnect {
    genLogFiles::logprint("disconnecting");
    print $socket "goodbye task[\"cliexit\"];\n";
    $socket->flush(); #74852
#    syswrite($socket,"goodbye task[\"cliexit\"];\n");
    getResponse($socket);
    $select->remove($socket);
    $socket->close();
}

#
# close the created Socket
#
sub closeSocket {

    if ($ipv6support && IPV6) {
       $socket_IPV6->free();
    }
    else {
       close($socket);
    }
}

#
# read Wellcom message from MM
#
sub getWellcome {
       my $line = '';

       #
       # IPv6 connection
       if ($ipv6support && IPV6) {
           $line = $socket_IPV6->readline();
       }
       #
       # IPv4 connection
       else {
          $line = <$socket>;
       }

       return $line;
}

sub getCommand {
    my $line = '';
    my $ch;
    my $rv;

    # in case we have a command saved in the meantime
    # we send it now before reading any new command from
    # the TCP/IP stream
    if (length($saveCommand) != 0) {
        $rv = $saveCommand;
        $saveCommand = "";
        return $rv;
    }
    return readCommand();
}

#
# waiting only for accept response
#
sub getAcceptedResponse {
    my $line = '';
    my $error = 0;

    do {
        $line = readCommand();
        if (length($line) != 0) {
            # remove leading and trailing space chars
            $line =~ s/^\s+//;
            $line =~ s/\s+$//;
            # check if we've got a command while waiting for a response
            if (!($line =~ /^\s*response\s+.*/)) {
                if (length($saveCommand) == 0) {
                    # save the command and wait again for response
                    $saveCommand = $line;
                    genLogFiles::logprint("saving command $line");
                }
                else {
                    genLogFiles::logprint("possibly loosing command $line");
                }
                $line = "";
            }
        }
        else {
            genLogFiles::logprint("error reading response:$!");
            exit 1;
        }
    } while (length($line) == 0);

    genLogFiles::logprint("got response =>$line<=");

    if ($line =~ /^\s*response\s+unacceptable.*/) {
        genLogFiles::logprint("command unacceptable: $line");
        doExit();
    }
}

#
# waiting for accept and success responses
#
sub getResponse {
    my $dontExitOnError = shift;
    my $line = '';
    my $error = 0;
    my $status;

    do {
        $line = readCommand();
        if (length($line) != 0) {
            # remove leading and trailing space chars
            $line =~ s/^\s+//;
            $line =~ s/\s+$//;
            # check if we've got a command while waiting for a response
            if (!($line =~ /^\s*response\s+.*/)) {
                if (length($saveCommand) == 0) {
                    # save the command and wait again for response
                    $saveCommand = $line;
                    genLogFiles::logprint("saving command $line");
                }
                else {
                    genLogFiles::logprint("possibly loosing command $line");
                }
                $line = "";
            }
        }
        else {
            genLogFiles::logprint("error reading response:$!");
            doExit();
        }
    } while (length($line) == 0);

    genLogFiles::logprint("got response =>$line<=");

    if ($line =~ /^\s*response\s+unacceptable.*/) {
        genLogFiles::logprint("command unacceptable: $line");
        doExit();
    }

    #
    # only wait for the next response if this was the accepted response
    #
    if ($line =~ /^\s*response\s+task\s*\[\s*(\".*\")\s*\]\s*accepted.*/) {
        do {
            $line = readCommand();
            if (length($line) != 0) {
                # remove leading and trailing space chars
                $line =~ s/^\s+//;
                $line =~ s/\s+$//;
                # check if we've got a command while waiting for a response
                # check if we've got a command while waiting for a response
                if (!($line =~ /^\s*response\s+.*/)) {
                    if (length($saveCommand) == 0) {
                        # save the command and wait again for response
                        $saveCommand = $line;
                        genLogFiles::logprint("saving command $line");
                    }
                    else {
                        genLogFiles::logprint("possibly loosing command $line");
                    }
                    $line = "";
                }
            }
            else {
                genLogFiles::logprint("error reading response:$!");
                doExit();
            }
        } while (length($line) == 0);

        genLogFiles::logprint("got response =>$line<=");
    }
    #
    # remove trailing ;
    #
    $line =~ s/;$//;

    $status = getResponseStatus($line);

    if (defined $dontExitOnError && $dontExitOnError) {
        return $line;
    }

    if (lc($status) eq "unacceptable") {
        genLogFiles::logprint("command unacceptable: $line");;
        doExit();
    }
    elsif (lc($status) eq "cancelled") {
        genLogFiles::logprint("command cancelled: $line");
        doExit();
    }
    elsif (lc($status) eq "error") {
        my $errorCode = getResponseErrorCode($line);
        genLogFiles::logprint("error executing command:$line");
        doExit();
    }

    return $line;
}

sub getTaskId {
    my $line = shift;

    $line =~ /^\s*response\s+task\s*\[\"([^\"]*)\"\].*/;
    return $1;
}

sub getResponseStatus {
    my $line = shift;

    if ($line =~ /^\s*response\s+unacceptable.*/) {
        return "unacceptable";
    }
    $line =~ /^\s*response\s+task\s*\[\"([^\"]*)\"\]\s*([a-zA-Z]+).*/;
    return $2;
}

sub getResponseErrorCode {
    my $line = shift;

    $line =~ /^\s*response\s+task\s*\[\s*\"([^\"]*)\"\s*\]\s*error\s*\[\s*\"([^\"]*)\"\s*\"([^\"]*)\"\s*\].*/;
    return "$3";
}

sub getKeyValueResults {
    my $line = shift;
    my @values;
    my $i = 0;
    my @objs = (split /\btext/, $line);

    for ($i = 1; $i <= $#objs; $i++) {
        $_ = $objs[$i];
        # remove leading and trailing space chars
        s/^\s+//;
        s/\s+$//;
        # remove leading and trailing brackets
        s/^\[//;
        s/\]$//;
        # remove leading and trailing quote
        s/^\"//;
        s/\"$//;
        # split results at " "
        push @values, [ (split /\"\s+\"/) ];
    }
    return @values;
}

sub doExit {
    sleep($secondsBeforeExit);
    exit 1;
}

#
# Reading from IPV4 or IPV6 socket
#
sub readCommand {
    my $line = "";

    if ($ipv6support && IPV6) {
       $line = readCommand_IPV6();
    }
    else {
       $line = readCommand_IPV4();
    }
    return $line;
}

#
# Read the information from the IPv4 Socket
#
sub readCommand_IPV4 {
    my $line = "";
    my $rv = 0;
    my $ch;

    do {
        $line = "";
        do {
            if ($select->can_read(600)) {
                if (($rv = sysread($socket, $ch, 1)) < 1) {
                    return "";
                }
                if (!($ch eq "\n" || $ch eq "\r")) {
                    $line .= $ch;
                }
            }
            else {
		
                my $kstr = "keepalive;\n";
		$socket->flush(); #74852
		genLogFiles::logprint("info: sent keepalive message to TSLM server\n");
                if (syswrite($socket, "keepalive;\n") < 10) {
                    genLogFiles::logprint("error: socket not connected\n");
                    return "";
                }
		$socket->flush(); #74852
            }
        } while !($ch eq ";");
    } while ($line =~ /response\s+unacceptable.*/);
    return $line;
}

#
# Read the information from the IPv6 Socket
#
sub readCommand_IPV6 {
    my $line = "";
    my $rv = 0;
    my $ch;

    do {
        $line = "";
        do {
            if ($socket_IPV6->is_readable(600)) {
                if (($rv = $socket_IPV6->read($ch,1)) < 1) {
                    return "";
                }
                if (!($ch eq "\n" || $ch eq "\r")) {
                    $line .= $ch;
                }
            }
            else {
                if ($socket_IPV6->is_error) {
                    genLogFiles::logprint("error: socket not connected\n");
                    return "";
                }
            }
        } while !($ch eq ";");
    } while ($line =~ /response\s+unacceptable.*/);
    return $line;
}

#
# Resolve a DNS name entry in the configuration file for 'dmhost'.
#
sub getdmhostIp {

    my $dmhost = shift;
    my $ipv6support = shift;

    my ($sock, $ai_ref, $ipaddr);
    my @res   = ();


    if ($dmhost) {
       #
       # IPv6 connection
       if ($ipv6support && IPV6) {

          $sock = Socket::Class->new();
          @res  = $sock->getaddrinfo($dmhost,0,'inet6','0','STREAM');

          if ($sock->is_error ||
              $sock->error)   {
             genLogFiles::logprint("Function getaddrinfo returned an error for address lookup of dmhost \'$dmhost\': ".$sock->error.".");
             genLogFiles::logprint("Invalid dmhost \'$dmhost\' entered in configuration file.");
             $dmhost = "";
          }
          else {
             $ai_ref = shift @res;
             $dmhost = $ai_ref->{addr};
          }
          $sock->free;
       }
       #
       # IPv4 connection
       else {

          $ipaddr = inet_aton($dmhost);
          if ($ipaddr) {
             $dmhost = inet_ntoa($ipaddr);
          }
          else {
             genLogFiles::logprint("Invalid dmhost \'$dmhost\' entered in configuration file.");
             $dmhost = "";
          }
       }
    }

    return $dmhost;
}

#
# Reorder the list of IP addresses in the configuration file.
# The last reachable IP will be put in front of the list.
# In case there is a problem reading or writing the file
# in this function we do not exit the program.
#
# $input = <Active IP>:<Active Port>
# $hostvalue = Entire value of $host (list of IPs)
# $portvalue = Value of $port
# $configFile = Configuration file used
#
sub reorderIPs() {
    my ($activeIp, $activePort, $hostvalue, $portvalue, $configFile) = @_;
    my $index=0;
    my $update=0;
    my @ipAddresses = ();
    my @lines = ();
    my $IPAddress;
    my $IPPort;
    my $newIPList="";
    my $ret;


    @ipAddresses = split(/,/,$hostvalue);
    foreach my $currentIp (@ipAddresses) {

        $IPAddress = "";
        $IPPort    = "";

        #Extract the port information.
        ($IPAddress, $IPPort) = parsingIpAddress($currentIp);

        # remove any leading/trailing whitespace
        $IPAddress =~ s/^\s*//;
        $IPAddress =~ s/\s*$//;

        if ($IPPort eq "" || (!defined $IPPort) ) {
            if (($IPAddress eq $activeIp) && ($portvalue eq $activePort)) { last };
        }
        else {
            $IPPort =~ s/^\s*//;
            $IPPort =~ s/\s*$//;
            if (($IPAddress eq $activeIp) && ($IPPort eq $activePort)) { last };
        }
        $index++;
    }

    # switch the IP addresses in the list
    my $switchIp = $ipAddresses[0];
    if ($IPPort eq "" || (!defined $IPPort)) {
        $ipAddresses[0] = $activeIp;
    }
    else {
        # IPv6 IP-address in brackets
        if ($ipv6support && IPV6) {
           $ipAddresses[0] = "[".$activeIp."]:".$activePort;
        }
        else {
           $ipAddresses[0] = $activeIp.":".$activePort;
        }
    }
    $ipAddresses[$index] = $switchIp;

    # Update the IP address list
    $newIPList = join (',',@ipAddresses);

    # Reorder the list of IP addresses in the configuration file
    # (if the file $configFile exists).
    if (! -e $configFile) {
        genLogFiles::logprint("Configuration file $configFile not found.");
        return 1;
    }

    $ret = open(CONFFILE, $configFile);
    if ($ret==0 || !defined $ret) {
        genLogFiles::logprint("Can not open configuration file $configFile");
        return 1;
    }

    @lines = <CONFFILE>;

    $ret = close(CONFFILE);
    if (!$ret || !defined $ret) {
        genLogFiles::logprint("Can not close configuration file $configFile");
        return 1;
    }

    # Create a "new" configuration file with the modified host list
    $ret = open(CONFFILE, ">$configFile");
    if ($ret==0 || !defined $ret) {
        genLogFiles::logprint("Can not open configuration file $configFile");
        return 1;
    }

    foreach my $line (@lines) {
        # Check if the string in the config file is of the format
        # VARIABLE=VALUE or VARIABLE="VALUE".
        if($line =~ m/^\s*(\w+)\s*=\s*((?:\".*?\")|(?:[^\"]*?))\s*$/g) {
            $var = $1;
            $val = $2;

            if($var =~ m/^\s*SERVER\s*$/i) {
                # Update the list
                $line = $var."=".$newIPList."\n";
                $update = 1;
            }
        }

        # Write changed information back to file
        print CONFFILE $line;
    }

    $ret = close(CONFFILE);
    if (!$ret || !defined $ret) {
        genLogFiles::logprint("Can not close configuration file $configFile");
        return 1;
    }

    if ($update) {
        genLogFiles::logprint("Reordered the list of IP addresses in configuration file. New primary IP address is $ipAddresses[0].");
        return 0;
    }
    else {
        genLogFiles::logprint("Could not reorder the list of IP addresses in configuration file.");
        return 1;
    }
}

#
# Parsing the IP-Address for IPv4 and IPv6.
#
sub parsingIpAddress {

    my $inputIpAddress = shift;
    my $defaultPort    = shift;
    my ($IPAddress, $IPPort);


    # IPv6 Address in brackets with port (e.g. [2008::1]:6740)
    if ( index($inputIpAddress,']') > 0 ) {
       ($IPAddress,$IPPort) = split(']',$inputIpAddress);

       $IPAddress =~ s/\[//;          # Remove existing '[' from IPAdress
       $IPPort    =~ s/\://;          # Remove existing ':' from Port
    }
    # IPv6 Address without port, multiple ':' (e.g. ::1)
    elsif ( index($inputIpAddress,':') < rindex($inputIpAddress,':') ) {
       $IPAddress = $inputIpAddress;
       $IPPort    = '';
    }
    # IPv4 Address
    else {
       # Extract the port information (if specified). If none is given,
       # the port information in $Ermm::CLI::Conf::port will be used.
       ($IPAddress, $IPPort) = split(':', $inputIpAddress);
    }

    # Remove any leading/trailing whitespace
    $IPAddress =~ s/^\s*//;
    $IPAddress =~ s/\s*$//;

    if ($IPPort eq "" || (!defined $IPPort)) {
       $IPPort = $defaultPort;
    }
    else {
       $IPPort =~ s/^\s*//;
       $IPPort =~ s/\s*$//;
    }

    return $IPAddress, $IPPort;
}

# Read the first parameter as string and return the base64 coded string.
#
# Summary (Details are in RF2045)
# - Pad input stream with value 0 to have number of chars divisible by 3
# - Treat each 3 chars as 24 bit value
# - Split 24 bit value in 4 values with 6 bit each
# - Encode each 6 bit value with the chars A-Za-z+/
# - Replace the padded bytes in the input stream with '='
sub base64_encode {
  my ($input) = (@_);

  my @inarray = split//, $input;

  # Input padding: Add the value '0' until the number of
  # bytes can be divided by 3 and remember the number of
  # padding bytes.
  my $padding = 0;
  while(($#inarray+1) % 3 != 0) {
    push @inarray, chr 0;
    $padding+= 1;
  }

  my @outarray;
  for(my $i=0; $i < $#inarray; $i+=3) {
    my $bits;
    $bits += (ord $inarray[$i+0]) << 16;
    $bits += (ord $inarray[$i+1]) <<  8;
    $bits += (ord $inarray[$i+2]);

    push @outarray, ($bits & 0xFC0000) >> 18;
    push @outarray, ($bits & 0x03F000) >> 12;
    push @outarray, ($bits & 0x000FC0) >>  6;
    push @outarray, ($bits & 0x00003F);
  }

  foreach(@outarray) {
    if( (0 <= $_) and ($_ <= 25) ) {
      $_ = chr ($_ + ord 'A'); # A-Z

    } elsif( (26 <= $_) and ($_ <= 51) ) {
      $_ = chr ($_ - 26 + ord 'a'); # a-z

    } elsif( (52 <= $_) and ($_ <= 61) ) {
      $_ = chr ($_ - 52 + ord '0'); # 0-9

    } elsif($_ == 62) {
      $_ = '+';

    } elsif($_ == 63) {
      $_ = '/';
    }
  }

  $outarray[-2] = '=' if $padding == 2;
  $outarray[-1] = '=' if $padding >= 1;

  return join '', @outarray;
}

sub L10N_init {
   my ($lng, $cntry) = @_;

   unless ($lh = nls::hdm_L10N -> get_handle("hdm\_$lng\_$cntry")) {
      if (($lng eq 'en') && ($cntry eq 'us')) {
         genLogFiles::logprint("Error loading default NLS module hdm_en_us!");
         doExit();
      }
      else {
         print "Error loading NLS module hdm\_$lng\_$cntry! Loading default NLS module (en_US)...\n\n";
         genLogFiles::logprint("Error loading NLS module hdm\_$lng\_$cntry! Loading default NLS module (en_US)...");
         $lh = nls::hdm_L10N -> get_handle('hdm_en_us') || die "Error loading default NLS module hdm_en_us!\n";
      }
   }
   $lh -> fail_with("lookup_failure_handler");
}


sub printtext (@) {
   print( $lh -> maketext(@_) );
 }


sub gettext (@) {
   $lh -> maketext(@_);
 }

1;
