﻿#!/usr/bin/perl
#-----------------------------------------------------------------------
# Licensed Materials - Property of IBM
#
# IBM Tape System Library Manager
#
# (C) Copyright IBM Corp. 2004, 2018 All Rights Reserved
#-----------------------------------------------------------------------
# @version 1.14 - 18/05/14 18:56:24
#
# @brief
#
#  72568: Support Multi-threading in HDM
#  74566: Added EMM3010E
#  81050: Bugfix
#  81486: Bugfix
#  82656: Support release command
#  84765: Multi Instances
#  85061: bug fixes for write protect cartridge
#-----------------------------------------------------------------------

package GenericTools;
use 5.005;
use strict;
use POSIX;
use FindBin;
use genLogFiles;

use Math::BigInt; # for 64 bit values in arrayToHex

use constant MULTITHREADED => do{ eval { require threads }};

our $TSLM_HDMLIBDIR=$FindBin::Bin;
our $TSLM_LIBDIR=substr($TSLM_HDMLIBDIR,0,rindex($TSLM_HDMLIBDIR,"/"));
our $TSLM_BASEDIR=substr($TSLM_LIBDIR,0,rindex($TSLM_LIBDIR,"/"));;
our $TSLM_BINDIR=$TSLM_BASEDIR."/bin";
our $TSLM_CONFDIR=$TSLM_BASEDIR."/conf";
if ( ! -e $TSLM_BINDIR."/ermmGenHDM" ){
    $TSLM_BASEDIR="/opt/IBM/ermm";
    $TSLM_BINDIR="/opt/IBM/ermm/bin";
    $TSLM_LIBDIR="/opt/IBM/ermm/lib";
    $TSLM_HDMLIBDIR="/opt/IBM/ermm/lib/hdm";
    $TSLM_CONFDIR="/opt/IBM/ermm/conf";
}

#
# configuration variables
#
my $tapeutil = $TSLM_BINDIR."/ermmTapeutil";
if ($^O =~ /.*win32.*/i) {
    $tapeutil = "c:\\IBM\\ermm\\bin\\ermmTapeutil.exe";
}

# IP address of the HDM. This is used to indicate the HDM in
# error messages.
my $dmhost = '';

# Switches and list for getting logsense pages
my $getLogsenseData = 1;
my $useLogPage0 = 1;
my @pagesFromConfigfile;

# Switch for enabling/disabling cartridge pcl checking
my $checkLoadedPCL = 0;
my $loadedPCL = '';

# Switch for enabling/disabling drive snr checking
my $checkDriveSNR = 0;

my $retryCount = 32;
my $secondsToSleepBeforeRetry = 5;

#############################################################################
# don't change any code below this line!
#############################################################################
sub getIBMTapeDevices {
    my $ignoreableDevices = shift;
    my($kernel, $hostname, $release, $version, $hardware) = uname();

    if ($kernel =~ /linux.*/i) {
        return getLinuxIBMTapeDevices($ignoreableDevices);
    }
    else {
        return getAIXIBMTapeDevices($ignoreableDevices);
    }
}

#
# create an array which contains all IBM tape devices
#
sub getLinuxIBMTapeDevices {
    my $p = shift;
    my @ignore = @$p;
    my @ibm;
    my ($line, $id, $model, $sn);
    my @rmts = `cat /proc/scsi/IBMtape`;
    my $rec = {};

    foreach $line (@rmts) {
        my $rec = {};
        my $val;

        chomp $line;
        if (($id, $model, $sn) =
            ($line =~ /^([0-9]+)\s+(\S+)\s+(\S+).*/)) {
            if (!belongsTo(\@ignore, "/dev/IBMtape$1")) {
                $rec->{"Handle"} = "/dev/IBMtape$1";
                $rec->{"Manufacturer"} = "IBM";
                $rec->{"Model"} = $model;
                $rec->{"SerialNumber"} = $sn;
                $rec->{"Driver"} = "ATAPE";
                push @ibm, $rec;
            }
        }
    }

    return @ibm;
}

#
# create an array which contains all IBM tape devices
#
sub getAIXIBMTapeDevices {
    my $p = shift;
    my @ignore = @$p;
    my @ibm;
    my @rmts = `lsdev -c tape | grep Available | grep rmt | grep \"IBM\" | awk \'{print \$1}\'`;
    my $rec = {};

    foreach (@rmts) {
        my @res = `lscfg -vl $_`;
        my $rec = {};
        my $val;
        my $dev;

        chomp;
        $dev = "/dev/$_";

        if (!belongsTo(\@ignore, $dev)) {
            $rec->{"Handle"} = $dev;
            foreach (@res) {
                if ($_ =~ /.+\-W([a-zA-Z0-9\.\-]+)\-.*$/) {
                    $rec->{"WWN"} = $1;
                }
                elsif ($_ =~ /.+Manufacturer\.{2,}([a-zA-Z0-9\.\-]+)\s*$/) {
                    $rec->{"Manufacturer"} = $1;
                }
                elsif ($_ =~ /.+Model\.{2,}([a-zA-Z0-9\.\-]+)\s*$/) {
                    $rec->{"Model"} = $1;
                }
                elsif ($_ =~ /.+Number\.{2,}([a-zA-Z0-9\.\-]+)\s*$/) {
                    $rec->{"SerialNumber"} = $1;
                }
            }
            $rec->{"Driver"} = "ATAPE";
            push @ibm, $rec;
        }
    }

    return @ibm;
}

#
# check if a string passed as 2nd parameter is member of array
# passed as 1st parameter
#
sub belongsTo {
    my $p = shift;
    my $s = shift;
    my @a = @$p;

    foreach (@a) {
        if ($_ eq $s) {
            return 1;
        }
    }
    return 0;
}

#
# start tapeutil load command in separate thread
#
sub Load {
    my $handle = shift;
    my $taskId = shift;
    my $cartpcl = shift;
    my $usethreads = shift;

    if ($usethreads && MULTITHREADED) {
	genLogFiles::logprint("creating new thread for load command");
        my $thread = threads->create(\&LoadThread, $handle, $taskId, $cartpcl);

        if (defined($thread)) {                     # @MR2
            $thread->detach();
        } else {                                    # @MR2
            genLogFiles::logprint("Faild to create the new thread for load. Executing in main thread as normal."); # @MR2
            LoadThread($handle, $taskId, $cartpcl); # @MR2
        }                                           # @MR2
    }
    else {
        LoadThread($handle, $taskId, $cartpcl);
    }
}

sub LoadThread {
    my $handle = shift;
    my $taskId = shift;
    my $cartpcl = shift;

    if (MULTITHREADED) {                                       # @MR2
        my $thr = threads->self();                             # @MR2
        my $tid = $thr->tid();                                 # @MR2
        genLogFiles::logprint("($tid): begin of LoadThread."); # @MR2
    }                                                          # @MR2

    $cartpcl = substr($cartpcl,0,6);

    my $util;
    if ($handle->{"TapeUtil"}) {
        $util = $handle->{"TapeUtil"};
    }
    else {
        $util = $tapeutil;
    }

    if(Execute('load', $util, $handle, $taskId)) {
	#
	# check loaded pcl if load was successful
	#
	if ( $checkLoadedPCL == 1 ) {
            my $chkPCLparms =  getLoadParmsForCheckPCL($handle);
	    #
	    # check only suported drives (IBM)
	    #
            if ( $chkPCLparms ne "nocheck") {
                $chkPCLparms = $chkPCLparms . " -p " .  $cartpcl;
                $loadedPCL = "";
		# give the library a little bit time to write the pcl into the cart memory 
		# if this is a new cartridge
		sleep(3);
		#
		# if check was successful return success to MM
		# else unload just loaded cartridge and return EMM3301E to MM
		# in case of an error, the real loaded pcl is returned in var $loadedPCL
		#
                if(Execute('checkpcl', $util, $handle, $taskId, $chkPCLparms)) {
		    ErmmHdm::send("response task[$taskId] success;\n");
                }
                else {
		    Execute('unload', $util, $handle, $taskId);
		    ErmmHdm::send(createErrorResponse($taskId, 'EMM3301E',
                                                      'dmhost', $dmhost,
						      'epcl', $cartpcl, 'lpcl',
                                                      $loadedPCL));
                }
            }
            else {
                ErmmHdm::send("response task[$taskId] success;\n");
            }
	}
	else {
	    ErmmHdm::send("response task[$taskId] success;\n");
	}
    }

    if (MULTITHREADED) {                                     # @MR2
        my $thr = threads->self();                           # @MR2
        my $tid = $thr->tid();                               # @MR2
        genLogFiles::logprint("($tid): end of LoadThread."); # @MR2
    }                                                        # @MR2
}

#
# start tapeutil unload command in separate thread
#
sub Unload {
    my ($handle, $taskId, $skippages, $codelist, $usethreads) = @_;

    if ($usethreads && MULTITHREADED) {
	genLogFiles::logprint("creating new thread for unload command");
        my $thread = threads->create(\&UnloadThread, $handle, $taskId, 
                                     $skippages, $codelist);
        if (defined($thread)) {                                    # @MR2
            $thread->detach();
        } else {                                                   # @MR2
            genLogFiles::logprint("Faild to create the new thread for unload. Executing in main thread as normal."); # @MR2
            UnloadThread($handle, $taskId, $skippages, $codelist); # @MR2
        }                                                          # @MR2
    }
    else {
        UnloadThread($handle, $taskId, $skippages, $codelist);
    }
}

sub UnloadThread {
    my ($handle, $taskId, $skippages, $codelist) = @_;

    if (MULTITHREADED) {                                         # @MR2
        my $thr = threads->self();                               # @MR2
        my $tid = $thr->tid();                                   # @MR2
        genLogFiles::logprint("($tid): begin of UnloadThread."); # @MR2
    }                                                            # @MR2

    my $util;
    if ($handle->{"TapeUtil"}) {
        $util = $handle->{"TapeUtil"};
    }
    else {
        $util = $tapeutil;
    }

    if (Execute('unload', $util, $handle, $taskId)) {
        ErmmHdm::send("response task[$taskId] success "
                      . GenericTools::GetLogSense($handle, $skippages, $codelist)
                      . ";\n");
    }

    if (MULTITHREADED) {                                       # @MR2
        my $thr = threads->self();                             # @MR2
        my $tid = $thr->tid();                                 # @MR2
        genLogFiles::logprint("($tid): end of UnloadThread."); # @MR2
    }                                                          # @MR2
}

sub TestUnitReady {
  my $handle = shift;
  my $taskId = shift;
  my $drive = shift;
  my $util;
  if ($handle->{"TapeUtil"}) {
      $util = $handle->{"TapeUtil"};
  }
  else {
      $util = $tapeutil;
  }
  genLogFiles::logprint("Failures during tur probes are expected");

  return(Execute('tur', $util, $handle, $taskId));
}

#
# start tapeutil inquiry command in separate thread
#
sub Inquiry {
    my $handle = shift;
    my $taskId = shift;
    my $drive = shift;

    InquiryThread($handle, $taskId, $drive);
}

sub InquiryThread {
    my $handle = shift;
    my $taskId = shift;
    my $drive = shift;
    my $util;
    if ($handle->{"TapeUtil"}) {
        $util = $handle->{"TapeUtil"};
    }
    else {
        $util = $tapeutil;
    }

    if (Execute('inquiry', $util, $handle, $taskId)) {
        ErmmHdm::send("ready task[\"ready\"] drive[\"$drive\"];\n");
        ErmmHdm::send("response task[$taskId] success;\n");
    }
    else {
        ErmmHdm::send("ready task[\"ready\"] drive[\"$drive\"] not;\n");
    }
}

#
# start tapeutil erase command in separate thread
#
sub Erase {
    my $handle = shift;
    my $taskId = shift;
    my $usethreads = shift;

    if ($usethreads && MULTITHREADED) {
	genLogFiles::logprint("creating new thread for erase command"); # @MR2
        my $thread = threads->create(\&EraseThread, $handle, $taskId);

        if (defined($thread)) {            # @MR2
            $thread->detach();
        } else {                           # @MR2
            genLogFiles::logprint("Faild to create the new thread for erase. Executing in main thread as normal."); # @MR2
            EraseThread($handle, $taskId); # @MR2
        }                                  # @MR2
    }
    else {
        EraseThread($handle, $taskId);
    }
}

sub EraseThread {
    my $handle = shift;
    my $taskId = shift;

    if (MULTITHREADED) {                                        # @MR2
        my $thr = threads->self();                              # @MR2
        my $tid = $thr->tid();                                  # @MR2
        genLogFiles::logprint("($tid): begin of EraseThread."); # @MR2
    }                                                           # @MR2

    my $util;
    if ($handle->{"TapeUtil"}) {
        $util = $handle->{"TapeUtil"};
    }
    else {
        $util = $tapeutil;
    }

    if (Execute('erase', $util, $handle, $taskId)) {
        ErmmHdm::send("response task[$taskId] success;\n");
    }

    if (MULTITHREADED) {                                      # @MR2
        my $thr = threads->self();                            # @MR2
        my $tid = $thr->tid();                                # @MR2
        genLogFiles::logprint("($tid): end of EraseThread."); # @MR2
    }                                                         # @MR2
}

#
# Release scsi reservation
#
sub Release {
    my $handle = shift;
    my $taskId = shift;
    my $usethreads = shift;

    if ($usethreads && MULTITHREADED) {
	genLogFiles::logprint("creating new thread for release command");
        my $thread = threads->create(\&ReleaseThread, $handle, $taskId);

        if (defined($thread)) {
            $thread->detach();
        } else {
            genLogFiles::logprint("Faild to create the new thread for release. Executing in main thread as normal."); # @MR2
            ReleaseThread($handle, $taskId);
        }
    }
    else {
        ReleaseThread($handle, $taskId);
    }
}

sub ReleaseThread {
    my $handle = shift;
    my $taskId = shift;

    if (MULTITHREADED) {
        my $thr = threads->self();
        my $tid = $thr->tid();
        genLogFiles::logprint("($tid): begin of ReleaseThread.");
    }
    my $util;
    if ($handle->{"TapeUtil"}) {
        $util = $handle->{"TapeUtil"};
    }
    else {
        $util = $tapeutil;
    }

    if(Execute('release', $util, $handle, $taskId)) {            
	ErmmHdm::send("response task[$taskId] success;\n");
    }

    if (MULTITHREADED) {
        my $thr = threads->self();
        my $tid = $thr->tid();
        genLogFiles::logprint("($tid): end of ReleaseThread.");
    }
}

#
# Execute SCSI command by running the ermmXXXutil backend.
#
# This function returns 1 on success, in this case the caller has to proceed with the protocol
#   handling and send 'response success' to the Media Manager.
#
# If something went wrong, the function returns 0 and automatically sends 'response error' to
#   the Media Manager. In this case the caller must not send a response to the Media Manager.
#
# Note: Yes, this behaviour is strange and should be improved in the future.
#
sub Execute {
    my ($command, $util, $handle, $taskId, $chkparms) = @_;

    my $tid = "m";                      # @MR2
    if (MULTITHREADED) {                # @MR2
        my $thr = threads->self();      # @MR2
        $tid = $thr->tid();             # @MR2
    }                                   # @MR2

    my $cmdString = "";

    if ( $checkDriveSNR == 1 ) {
	$cmdString = qq/$util -f $handle->{"Device"} -h $handle->{"HostId"} -s $handle->{"ScsiId"} /
	    . qq/-b $handle->{"BusId"} -l $handle->{"LunId"} -n $handle->{"SerialNumber"} $command/;
    }
    else {
	$cmdString = qq/$util -f $handle->{"Device"} -h $handle->{"HostId"} -s $handle->{"ScsiId"} /
	    . qq/-b $handle->{"BusId"} -l $handle->{"LunId"} $command/;
    }


    my $serial = $handle->{"SerialNumber"};
    my $driveModel = $handle->{"Model"};

    if($command eq 'checkpcl') {
	$cmdString = qq/$util -f $handle->{"Device"} -h $handle->{"HostId"} -s $handle->{"ScsiId"} /
	    . qq/-b $handle->{"BusId"} -l $handle->{"LunId"} $chkparms $command/;
    }

    my $retries = $retryCount;
    if ($command eq 'release'){
        $retries = 1;
    }

    # Defined here, because it is referenced behind the retry loop
    my $errorNumber;
    my $errorString;
    while ($retries-- > 0) {
        genLogFiles::logprint("($tid): Issuing command: $cmdString"); # @MR2

        my $rc = open EXECHANDLE, "$cmdString|";

        if(!$rc) {
            genLogFiles::logprint("($tid): failed to execute $!"); # @MR2
            ErmmHdm::send(createErrorResponse($taskId, 'EMM3004E',
                                              'tapeutil', $util,
                                              'cmd', $command,
                                              'dmhost', $dmhost));
            return 0;
        }

        my ($success, $key, $asc, $ascq, $hexstring, $serialFound);
        $errorNumber = 0;
        $errorString = "";
        while(<EXECHANDLE>) {
            chomp;
            genLogFiles::logprint("($tid): From $util: $_"); # @MR2

            # Strategy: Run the program, remember the messages received
            #   (should be only one, but who knows) and evaluate the status later.
            if(/^Success:/) {
                $success = 1;

            }
            elsif(/^System Error: (\d+) (.+)$/) {
                ($errorNumber, $errorString) = ($1, $2);
                # System independent newline removal at the end of the line
                $errorString =~ s/[\012\015]*$//;
            }
            elsif(/^SCSI Sense Error: (.+)$/) {
                $hexstring = $1;

                ($key, $asc, $ascq) = parseSenseBuffer($hexstring);
            }
            elsif(/^Serialnumber Check Error: (.*)$/) {
                $serialFound = $1;
            }
        }

        close EXECHANDLE;

        # Evaluate status here to be able to issue retries.

        # First, look at errors from the tape drive and issue retries or
        # follow other special cases. This is done, because after an sense
        # status from the drive, ermmTapeutil can still print other messages.

        # Go back and retry the test unit ready command again, if the
        # drive reports a unit attention
        if($command eq 'tur' && defined $key && $key==0x6) {
            sleep($secondsToSleepBeforeRetry);
            next;
        }

        # 0x2, 0x4, 0x1 == "becoming ready" Retry the load command in
        # this case, the drive will return a different status
        # when the load is complete (or has failed).
        if($command eq 'load' && defined $ascq
           && $key==0x2 && $asc==0x4 && $ascq==0x1) {
            sleep($secondsToSleepBeforeRetry);
            next;
        }

        # 0x2, 0x4, 0x3: This is reported from the IBM 3590 drive
        # for a load command to indicate that the load command only
        # works with an ACF (some kind of mini cartridge loader). So
        # return ok status if this condition is discovered for this drive.
        if($command eq 'load' && defined $ascq
           && $key==0x2 && $asc==0x4 && $ascq==0x3
           && $driveModel =~ /3590/) {
          return 1;
        }

        # STK 9840 report this condition, indicating that the drive is
        # still loading. So, try again.
        if($command eq 'load' && defined $ascq
           && $key==0x6 && $asc==0x28 && $ascq==0x0) {
            sleep($secondsToSleepBeforeRetry);
            next;
        }

        # All SCSI sense keys with code 2 have the same meaning: Not ready.
        # This will regularly mean that the unmount is in progress or
        # the drive might become available later.
        # Since we use the tur to check the communication with the drive
        # a not ready response confirms that we can communicate with the
        # device.
        # All libraries will wait for an operation to complete before they
        # complain about a move media so we can just ignore the not ready
        # on unload. Any permanent not ready of the drive will raise an
        # error in the following move media.
        #
        # "Medium not present" == 0x2, 0x3a, 0x0  (expected state)
        # "Not ready, manual intervention required" == 0x2, 0x4, 0x3
        # ...
        if( ($command eq 'unload' || $command eq 'tur')
            && defined $key
            && ($key==0x2 ) ) {
            return 1;
        }

        # We will ignore some sense keys and assume the operation worked:
        # key 0: No sense
        # key 1: recovered error
        # key 6: unit attention
        if(defined $key && ($key==0x0 || $key==0x1 || $key==0x6) ) {
            return 1;
        }

        # If there is an error from the drive (sense data), report it
        # as message
        if(defined $key && defined $asc && defined $ascq && defined $hexstring) {

	    if ($key==0x4 && $asc==0x53 && $ascq==0x0){ # media load or eject failed
		ErmmHdm::sendMessage('EMM3010E', 'error',
				     'errno', $errorNumber,
				     'dmhost', $dmhost,
				     'SerialNumber', $serial, 'cmd', $command,
				     'key', $key, 'asc', $asc, 'ascq', $ascq,
				     'data', $hexstring);
	    }elsif ($key==0x4 && $asc==0x52 && $ascq==0x0){ # cartridge fault
		ErmmHdm::sendMessage('EMM3010E', 'error',
				     'errno', $errorNumber,
				     'dmhost', $dmhost,
				     'SerialNumber', $serial, 'cmd', $command,
				     'key', $key, 'asc', $asc, 'ascq', $ascq,
				     'data', $hexstring);
	    }elsif ($key==0x7){ # data protect
		ErmmHdm::sendMessage('EMM3010E', 'error',
				     'errno', $errorNumber,
				     'dmhost', $dmhost,
				     'SerialNumber', $serial, 'cmd', $command,
				     'key', $key, 'asc', $asc, 'ascq', $ascq,
				     'data', $hexstring);
	    }else{
		ErmmHdm::sendMessage('EMM3002E', 'error',
				     'errno', $errorNumber,
				     'dmhost', $dmhost,
				     'SerialNumber', $serial, 'cmd', $command,
				     'key', $key, 'asc', $asc, 'ascq', $ascq,
				     'data', $hexstring);
	      }
        }

        if($success) {
            return 1;
        }
	else {
	    if($command eq 'checkpcl') {
		$loadedPCL = $errorString;
		return 0;
	    }
	}

        # Error code 46 is often generated by ermmTapeutil (why?). So
        # far, this seems only to be relevant for load, so return
        # success for other commands

        # ermmIbmutil and ermmTsmutil can return the error code 46
        # (not ready). Ignore this condition on all commands, except load.
        if ($errorNumber == 46 && $command ne 'load') {
            return 1;
        }

        # This has been observed on AIX with tape drive that has been
        # disconnected from the SAN: The device driver returned errno
        # 5 (I/O error), but there is no sense information from the
        # tape drive. This means that the host cannot reach the tape
        # drive. So, try again, probably the SAN recovers again. If
        # not, the error will be reported.
        if( $errorNumber == 5 && !defined $key) {
            sleep($secondsToSleepBeforeRetry);
            next;
        }

        # Previously, for all cases where ermmTapeutil got sense
        # information from the tape drive, it returned error code 46.
        # Now it returns the "real" error code 5 (this is the case at
        # least on AIX).
        #
        # For now, this is the same recovery as error code 46. But
        # with the "real" error code, the drive sense data, the
        # operating system type and the name of the ermmXXXutil, more
        # detailed error handling can be implemented _here_.
        if ($errorNumber == 5 && $command ne 'load') {
            return 1;
        }

	# ermmIbmutil return code for TUR when the drive is not yet loaded
        if( ($command eq 'tur') && ($errorNumber == 123) ) {
            return 1;
        }

        # ermmTsmutil return code for TUR when the drive is not yet loaded
        if ($errorNumber == 21) {
            return 1;
        }

        if ($serialFound) {
            ErmmHdm::send(createErrorResponse($taskId, 'EMM3008E',
                                      'dmhost', $dmhost,
                                      'drive', $serial, 'found', $serialFound));
          return 0;
        }

        # only error codes EAGAIN, EBUSY, EINVAL and ESPIPE
        # are tolerated
        #
        # add ECONNREFUSED to the errors which are tolerated
        #
        if ($errorNumber != 11 && $errorNumber != 16
            && $errorNumber != 29 && $errorNumber != 22
            && $errorNumber != 79) {
	    GetStatusData($handle, $taskId, 0);

            if (defined $key && defined $asc && 
                defined $ascq && defined $hexstring) {
		if ($key==0x4 && $asc==0x53 && $ascq==0x0){
		    ErmmHdm::send(createErrorResponse($taskId, 'EMM3010E',
						      'errno', $errorNumber,
						      'dmhost', $dmhost,
						      'SerialNumber', $serial, 
						      'cmd', $command,
						      'key', $key, 'asc', $asc, 
						      'ascq', $ascq,
						      'data', $hexstring));
		}elsif ($key==0x4 && $asc==0x52 && $ascq==0x0){
		    ErmmHdm::send(createErrorResponse($taskId, 'EMM3010E',
						      'errno', $errorNumber,
						      'dmhost', $dmhost,
						      'SerialNumber', $serial, 
						      'cmd', $command,
						      'key', $key, 'asc', $asc, 
						      'ascq', $ascq,
						      'data', $hexstring));
		}elsif ($key==0x7){
		    ErmmHdm::send(createErrorResponse($taskId, 'EMM3010E',
						      'errno', $errorNumber,
						      'dmhost', $dmhost,
						      'SerialNumber', $serial, 
						      'cmd', $command,
						      'key', $key, 'asc', $asc, 
						      'ascq', $ascq,
						      'data', $hexstring));
		}else{

	            ErmmHdm::send(createErrorResponse($taskId, 'EMM3002E',
						      'errno', $errorNumber,
						      'dmhost', $dmhost,
						      'SerialNumber', $serial, 
						      'cmd', $command,
						      'key', $key, 'asc', $asc, 
						      'ascq', $ascq,
						      'data', $hexstring));
		}
            }else{
                ErmmHdm::send(createErrorResponse($taskId, 'EMM3005E',
                                                  'cmd', $command,
                                                  'drive', $serial,
                                                  'dmhost', $dmhost,
                                                  'errno', $errorNumber,
                                                  'description', $errorString));
            }
            return 0;
        }

        sleep($secondsToSleepBeforeRetry);
    }

    # The command failed and no more retries are left. So return with
    # an error condition.
    GetStatusData($handle, $taskId, 0);

    # If we have some information about the error return it.
    if($errorNumber && $errorString && $command ne 'release' ) { # 82656
        ErmmHdm::send(createErrorResponse($taskId, 'EMM3007E',
                                          'cmd', $command, 'drive', $serial,
                                          'dmhost', $dmhost, 'errno', $errorNumber,
                                          'description', $errorString));
        return 0;
    }elsif ($command eq 'release'){
        return 1;
    }

    # In the unlikely case we got no info about the problem just complain.
    ErmmHdm::send(createErrorResponse($taskId, 'EMM3006E',
                                     'cmd', $command, 'drive', $serial,
                                     'dmhost', $dmhost));
    return 0;
}

# Read one string (like '0x01 0x02 0x03') as SCSI sense buffer and return
# sense key, asc and ascq
sub parseSenseBuffer {
    my ($hexstring) = @_;

    my @parts = split / +/, $hexstring;

    my @sense;
    foreach(@parts) {
        push @sense, hex $_;
    }

    my $key = $sense[2] & 0x0F;
    my $asc = $sense[12];
    my $ascq = $sense[13];

    return ($key, $asc, $ascq);
}

sub createErrorResponse {
    my ($taskId, $messageId, @messageArguments) = @_;

    my $response = qq/response task[$taskId] error["error" "$messageId"] /;
    $response .= qq/message[id["IBM" "TSLM" "$messageId"] /;

    if(@messageArguments) {
      $response .= 'arguments [';
      foreach(@messageArguments) {
        $response .= qq/"$_" /;
      }
      $response .= ']';
    }
    $response .= qq/];\n/;

    return $response;
}

# Read one drive sense data.
# If successful, an MMP attribute cmd will be send to MM
# Mode = 0 no error response on attribute cmd to MM, query sense data 
# in case of a failed HDM/device cmd
# Mode = 1 error response on attribute cmd to MM, query sense data in case 
# of a getstauts cmd from MM 
sub GetStatusData {
    my ($handle, $taskId, $mode) = @_;

    my $util;
    my $errorNumber = 0;
    my $errorString = "";
    my $DriveSenseData = ""; 
    my $serial = $handle->{"SerialNumber"};

    if ($handle->{"TapeUtil"}) {
        $util = $handle->{"TapeUtil"};
    }
    else {
        $util = $tapeutil;
    }

    my $cmd = qq/$util -f $handle->{"Device"} -h $handle->{"HostId"} -s $handle->{"ScsiId"} /
      . qq/-b $handle->{"BusId"} -l $handle->{"LunId"} status/;

    my $rc = open UTILHANDLE, "$cmd|";

    # On failure, don't return any data.
    unless ($rc) {
        genLogFiles::logprint("Error starting process for ".
                              "REQUEST SENSE commands, return code $rc");
        if ($mode == 1) {
	    ErmmHdm::send(createErrorResponse($taskId, 'EMM3005E',
					      'cmd', $cmd, 'drive', $serial,
					      'dmhost', $dmhost, 'errno', 
                                              $rc, 'description', 
                                              "Error starting process"));
        }
        return 1;
    }

    while(<UTILHANDLE>) {
        chomp;
	if(/^DriveSenseData: (.*)$/) {
	    $DriveSenseData = $1;
	}
	elsif(/^System Error: (\d+) (.+)$/) {
	    ($errorNumber, $errorString) = ($1, $2);
	    # at the end of $errorString sometimes exists an
	    # invalid char or a \n. if we have a \n chomp helps.
	    # if we have an invalid char only chop helps.
	    # therefore we use both methods. worst case, the last
	    # character of the errorstring is removed
	    chomp($errorString);
	    chop($errorString);

	}
    }
    close UTILHANDLE;

    if($errorNumber) {
	genLogFiles::logprint("GetStatusData error return code=$errorNumber");
	if ($mode == 1) {
	    ErmmHdm::send(createErrorResponse($taskId, 'EMM3005E',
					      'cmd', $cmd, 'drive', $serial,
					      'dmhost', $dmhost, 'errno', 
                                              $errorNumber, 'description', 
                                              "Error requesting drive sense data"));
	}
        return 1;
    }
    elsif($DriveSenseData) {
	ErmmHdm::sendDriveSenseData($serial, $DriveSenseData);
        #
        # we must not receive the reponse here, because this function
        # may get called in a separate thread and reading bytes from
        # tcp/ip stream would concur with the main thread reading a
        # command or response which leads to crippled messages fro
        # both of them; since this function should not let the whole
        # operation fail anyhow, it's even not nec. to get the response
        #
    }
    else {
	if ($mode == 1) {
	    ErmmHdm::send(createErrorResponse($taskId, 'EMM3005E',
					      'cmd', $cmd, 'drive', $serial,
					      'dmhost', $dmhost, 'errno', 
                                              $errorNumber, 'description', 
                                              "Error requesting drive sense ".
                                              "data, got no valid return data"));
	} 
        return 1;
    }
    
    if ($mode == 1) {
        ErmmHdm::send("response task[$taskId] success;\n");
    }

    return 0;
}

# Read one log sense page.
# If successful, the page is returned as an array
# with one element per byte of the page
sub GetLogSensePage {
    my ($handle, $pageCode) = @_;

    my $util;
    if ($handle->{"TapeUtil"}) {
        $util = $handle->{"TapeUtil"};
    }
    else {
        $util = $tapeutil;
    }

    my $cmd = qq/$util -f $handle->{"Device"} -h $handle->{"HostId"} -s $handle->{"ScsiId"} /
      . qq/-b $handle->{"BusId"} -l $handle->{"LunId"} logpage $pageCode/;

    my $rc = open UTILHANDLE, "$cmd|";

    # On failure, don't return any data.
    unless ($rc) {
        genLogFiles::logprint("Error starting process for LOG SENSE commands, return code $rc");
        return ();
    }

    my @data;

    while(<UTILHANDLE>) {
        chomp;
        # Expected line format: 4 spaces, line index, data, "["
        if(/ {4}[0-9A-F]{4} - ([0-9A-F ]{40}) \[/) {

            # Split data (in $1) to blocks of 16 bits each
            my @blocks = split/ +/, $1;

            # Split 16 bit blocks to 8 bit parts and store the values in @data
            foreach (@blocks) {
                my $first = hex substr($_, 0, 2);
                my $second = hex substr($_, 2, 2);

                # push second only when it was part of the string
                push @data, $first;
                push @data, $second if length($_)>2;
            }
        }
    }

    close UTILHANDLE;

    return @data;
}

# Read log sense pages by issuing the 'ermmXXXutil logsense' call.
# Note: This is the method that is currently used.
#
# When the log sense reading mechanism is revisited, this should
# probably be integrated with the Execute function, since there is
# some duplicated code here.
#
# This method returns the complete 'text [...]' response parameter
# with the key value pairs or an empty string. This can be just inserted
# in the response to the Media Manager.
sub GetLogSense {
    my ($handle, $skippages, $codelist) = @_;

    if ( (!defined $codelist) || ($getLogsenseData == 0) ) {
      # No log sense config file was available, so it does not
      # make sense to retrieve data that will not be reported anyway.
      return "";
    }

    my @dataPage0;
    my @pages;
    if ($useLogPage0 == 1) {
        @dataPage0 = GetLogSensePage($handle, 0);
        if(@dataPage0 == 0) {
            genLogFiles::logprint("Could not read log sense page 0.");
            return "";
        }
        @pages = getValidLogSensePages(@dataPage0);
    }
    else {
        @pages = @pagesFromConfigfile;
    }

    my $driveVendor = $handle->{'Manufacturer'};
    my $driveModel = $handle->{'Model'};

    # Key attributes in IEEE1244 syntax, this is returned later
    my $cmdstring = "";

    genLogFiles::logprint("Reading log sense pages started.");
    foreach my $page (@pages) {
        # Skip page if it is marked as 'skippage' for the current drive
        if(skipPage($page, $skippages, $driveVendor, $driveModel)) {
            next;
        }

        my @dataPage = GetLogSensePage($handle, $page);
        if(@dataPage == 0) {
            genLogFiles::logprint("Could not read log sense page $page.");
            next;
        }

        # Read attribute-value relations from page data
        my $data = parseLogSense(\@dataPage, $codelist, $driveVendor, $driveModel);

        foreach my $key(keys %$data) {
            my $value = $data->{$key};
            $cmdstring .= qq/"$key" "$value" /;
        }
    }
    genLogFiles::logprint("Reading log sense pages ended.");

    my $cstr = $cmdstring;
    $cstr =~ s/^\s+//;
    if ($cstr eq "") {
        return "";
    }
    return "text [ ". $cmdstring . "]";
}

# Determine if a log sense page should be skipped:
# Parameters: Page to check
#             Reference to skippages array
#             Drive vendor to check
#             Drive model to check
sub skipPage {
    my ($page, $skippages, $driveVendor, $driveModel) = @_;

    foreach(@$skippages) {
        my $vendorregex = $_->{'vendorregex'};
        my $modelregex = $_->{'modelregex'};

        if($page == $_->{'page'}
           && $driveVendor =~ m/$vendorregex/
           && $driveModel =~ m/$modelregex/) {
            return 1;
        }
    }

    return 0;
}

# Input:  Data from log sense page 0
# Output: Array of available log sense pages
sub getValidLogSensePages  {
    my (@data) = @_;

    my @pages;

    for(my $i = 4; $i <= $#data; $i++) {
        if($data[$i] != 0) {
            # Ignore page 0, since the page list already has been
            # retrieved from page 0 and there is no more data
            push @pages, $data[$i];
        }
    }
    return @pages;
}

# Process log sense data
# Input: Array reference of log sense data of one page
# Return: Reference to hash with parameter codes as keys
sub parseLogSense {
    my ($data, $codelist, $driveVendor, $driveModel) = @_;

    my $logsense = {};

    my $page = $data->[0] & 0x3F;

    # Use real length (from array) instead of the one in bytes 2 and 3
    my $length = @$data;

    # pi=="parameter index", start index of current parameter
    my $pi=4;
    while($pi<$length) {
        my $parameterCode = arrayToHex($data, $pi+0, $pi+1);
        # Flags are in $data->[2]
        my $parameterLength = $data->[$pi+3];
        my $value = arrayToHex($data, $pi+4, $pi+4+$parameterLength-1);

        $logsense->{$parameterCode} = $value;

        $pi += 4;
        $pi += $parameterLength;
    }

    # logsense is a reference to hash with parameter codes as keys
    # Now take the data and get the parameter names to use for the current drive

    my $attributes = {};
    foreach my $code (keys %$logsense) {

        foreach my $entry (@$codelist) {
            my $vendorregex = $entry->{'vendorregex'};
            my $modelregex = $entry->{'modelregex'};

            if($page == $entry->{'page'}
               && $code == $entry->{'code'}
               && $driveVendor =~ m/$vendorregex/
               && $driveModel =~ m/$modelregex/) {

                my $attr = $entry->{'attribute'};
                my $value = $logsense->{$code};

                my $ival = $entry->{'ignorevalue'};
                if(defined $ival && $ival == $value) {
                  # Ignore this value, continue with next entry
                  next;
                }

                $attributes->{$attr} = $value;
            }
        }
    }

    return $attributes;
}

sub Exit {
}

sub setCheckLoadedPCL {
    $checkLoadedPCL = shift;
}

sub setCheckDriveSNR {
    $checkDriveSNR = shift;
}

sub setGetLogsenseData {
    $getLogsenseData = shift;
}

sub setUseLogPage0 {
    $useLogPage0 = shift;
}

sub setTapeUtilExecutable {
        my $arg = shift;
	  $tapeutil = "\"".$arg."\"";
}

sub setRetryCount {
    $retryCount = shift;
}

sub setDmHost {
    my ($newdmhost) = @_;

    $dmhost = $newdmhost;
}

# Read and parse hdmlogsense.conf file
# Returns two references to structures with data from this file:
#  attributes is a list of attributes. Each list entry is a hash with the keys
#             page, code vendorregex, modelregex, attribute
#  skippages is a list of attributes. Each list entry is a hash with the keys
#             page, vendorregex, modelregex
#
# The two references returned are undefined, if it was not possible to open
# the config file.
sub readPagesConfig {
    my ($file) = @_;

    # Array reference pointing to hash references
    # with keys
    #   page, code, vendorregex, modelregex, attribute
    # If a log sense attribute matches the names from "attribute" will be used
    # to report it to the Media Manager.
    my $attributes = [];

    # Array reference pointing to hash references
    # with keys
    #   page, vendorregex, modelregex
    # If a log sense page matches here, it has to be ignored and not retrieved from the drive
    my $skippages = [];

    my $rc = open(FILE, $file);
    if (!defined $rc) {
      genLogFiles::logprint("Could not open log sense file $file: $!");
      return (undef, undef);
    }

    my $lastpage = " ";
    while(<FILE>) {
        chomp;

        next if(/^\s*#/); # Ignore comment lines
        s/#.*$//; # Remove comments at end of line

        my @fields = split / +/;

        # Special syntax for skipping pages
        if(defined $fields[0] && $fields[0] =~ /skippage/ && defined $fields[3]) {
            push @$skippages, {'page'        => hex $fields[1],
                               'vendorregex' => $fields[2],
                               'modelregex'  => $fields[3] };

        } elsif(defined $fields[4]) {
            push @$attributes, {'page'        => hex $fields[0],
                                'code'        => hex $fields[1],
                                'vendorregex' => $fields[2],
                                'modelregex'  => $fields[3],
                                'attribute'   => $fields[4] };

            # Optional 'ignorevalue' keyword
            if(defined $fields[6] && $fields[5] eq 'ignorevalue') {
              $attributes->[-1]->{'ignorevalue'} = hex $fields[6];
            }

            # add all pages given from hdmlogsense.conf file to
            # pagesFromConfigfile array. will be used instead of logpage0
            if ($lastpage ne $fields[0]) {
                push @pagesFromConfigfile, hex $fields[0];
                $lastpage = $fields[0];
            }
        }
    }

    close FILE;
    return ($attributes, $skippages);
}

# Treat a part of an array as a set of 8 bit values and
# convert them to one number
# Input:  Reference to array, from index, to index
# Return: The value when reading the sequence as one number.
sub arrayToHex {
    my ($array, $from , $to) = @_;

    my @data = @{$array}[$from .. $to];

    my $value = Math::BigInt->new(0);
    foreach(@data) {
        $value *= 256;
        $value += $_;
    }

    return $value;
}

sub getLoadParmsForCheckPCL {
    my $handle = shift;

    my $str = "";

    # this parms are manufaturer and type depended and must be passed to the ermmTapeutil call cmd checkpcl
    # -o = offset of the volser in sense data structure (vendor part)
    # -c = offset of the byte where to check ASCII or EBCDIC volser (359x only). if not used set to -1
    # -t = checkbyte pattern to test. 32 = 0x20 = bit 5 (359x only). if not used set to -1
    # when no IBM drive then set string to "nocheck"


    if ( $handle->{"Manufacturer"} =~ /.*IBM.*/ ) {
	if ( $handle->{"Model"} =~ /.*3590.*/ ) {
	    $str = " -o 83 -c 80 -t 32 ";
        }
        elsif ( $handle->{"Model"} =~ /.*3592.*/ ) {
            $str = " -o 83 -c 80 -t 32 ";
	}
	elsif ( $handle->{"Model"} =~ /.*3580.*/ ) {
            $str = " -o 22 -c -1 -t -1 ";
	}
	elsif ( $handle->{"Model"} =~ /.*ULTRIUM.*/ ) {
            $str = " -o 22 -c -1 -t -1 ";
	}
        else {
            $str = "nocheck";
        }
    }
    else {
        $str = "nocheck";
    }
    return $str;
}

1;
